#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>

#ifdef __FreeBSD__
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#endif

#ifdef __sun
#include <sys/socket.h>
#include <sys/sockio.h>
#endif

#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <net/if.h>


#include "../prips/prips.h"
#include "ipsc.h"

#define DOT(x) (8*(x))+((x)-1)

void get_bitmap(struct network *n)
{
	int i, dots = 0;
	char bitcode = 'n';

	n->bitmap[DOT(1)] = '.';
	n->bitmap[DOT(2)] = '.';
	n->bitmap[DOT(3)] = '.';

	for(i = 0; i < 35; i++)
	{
		if(i == DOT(1) || i == DOT(2) || i == DOT(3))
		{
			dots++;
			continue;
		}
		
		if(i == (n->bits + dots))
			bitcode = 's';
		
		if(i == (n->bits + n->subnet_bits + dots))
			bitcode = 'h';
	
		n->bitmap[i] = bitcode;
	}
}

/* 
 * Address size - network bits will get the maximum number 
 * of bits that can be allocated for subnets.  If the subnet
 * bits specified are greater than that amount we have a 
 * problem.
 */
int incorrect_subnet_bits(struct network *n)
{
	return(n->subnet_bits > (32 - n->bits) ? 1 : 0);
}

/*
 * Set the basic network information based on information that should be
 * in the network structure that is passed in to the function. 
 */
void set_network_info(struct network *n)
{
	n->host_bits = 32 - (n->bits + n->subnet_bits);
        n->host_max = pow(2, n->host_bits);

	n->subnet_mask = ~(n->mask | set_bits_from_right(n->host_bits));
        n->subnet_max = pow(2, n->subnet_bits);
	n->subnet_bits_max = 32 - n->bits;

	n->cisco_wildcard = ~(n->mask | n->subnet_mask);
	
	n->host_network_lbound = (n->mask | n->subnet_mask) & n->host_addr;
	n->host_network_ubound = n->host_network_lbound + n->host_max -1;

	get_bitmap(n);
}

void fprint_general(FILE *f, const struct network *n)
{
        /* Network information */
        fprintf(f, "%-25s%-10c\n", "Network class:", n->class);
        fprintf(f, "%-25s%-10s\n", "Network mask:", denumberize(n->mask));
        fprintf(f, "%-25s%-10X\n", "Network mask (hex):", n->mask);

        /* Print the network address if we have it */
        if(n->host_addr)
        {
                fprintf(f, "%-25s%-10s\n", "Network address:",
                        denumberize((n->mask | n->subnet_mask) & n->host_addr));
        }

        /* Subnet information */
        fprintf(f, "%-25s%-10d\n", "Subnet bits:", n->subnet_bits);
        fprintf(f, "%-25s%-10d\n", "Max subnets:", n->subnet_max);
        fprintf(f, "%-25s%-10s\n", "Full subnet mask:", 
		denumberize(n->mask | n->subnet_mask));
        fprintf(f, "%-25s%-10X\n", "Full subnet mask (hex):", n->mask | n->subnet_mask);

        /* Host information */
        fprintf(f, "%-25s%-10d\n", "Host bits:", n->host_bits);
        fprintf(f, "%-25s%-10d\n", "Hosts per subnet:", n->host_max);

        /* Bit mask */
        fprintf(f, "%-25s%-10s\n", "Bit map:", n->bitmap);

	fprintf(f, "\n");
}

void fprint_subnets_hosts(FILE *f, const struct network *n)
{
        int i;
        int network = n->addr;

        for(i = 0; i < n->subnet_max; i++)
        {
                fprintf(f, "%-17s", denumberize(network));
                fprintf(f, "%-17s", denumberize((network + n->host_max) - 1));
                if(n->host_addr >= network && n->host_addr < (network + n->host_max))
                        fprintf(f, "*");
                fprintf(f, "\n");
                network += n->host_max;
        }

	fprintf(f, "\n");
}

void fprint_host(FILE *f, const struct network *n)
{
	int subnet_range_ubound = n->addr + (n->subnet_max * n->host_max) -1;

	fprintf(f, "%-25s%-10s\n", "IP address:", denumberize(n->host_addr));
	fprintf(f, "%-25s%-10X\n", "Hexadecimal IP address:", n->host_addr);
	
	fprintf(f, "%-25s%-10s", "Address allocation range:",
		denumberize(n->addr));
	fprintf(f, " - %s\n", denumberize(subnet_range_ubound));

	fprintf(f, "%-25s%-10s\n", "Subnet mask:", denumberize(n->subnet_mask));
	fprintf(f, "%-25s%-10s\n", "Subnet ID:", denumberize(n->subnet_mask & n->host_addr));
	fprintf(f, "%-25s%-10s\n", "Network ID:", denumberize(n->mask & n->host_addr));
	fprintf(f, "%-25s%-10s\n", "Host ID:", 
		denumberize(~(n->mask | n->subnet_mask) & n->host_addr));
	fprintf(f, "\n");
}

void fprint_cidr(FILE *f, const struct network *n)
{
	fprintf(f, "%-25s%-10s\n", "Cisco Wildcard", denumberize(n->cisco_wildcard));
	fprintf(f, "\n");
}

void *get_interface_count(void *buffer, int *buf_len, int *sock)
{
	int sockfd, len, last_len, ifreq_size = sizeof(struct ifreq);
	struct ifconf ifc;	
	void *buf;

	if((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		perror("socket");
		exit(1);
	}

	/* Algorithm adapted from Stevens UNP. 
	 * Since we are not sure how many interfaces the system has, 
	 * we don't know what size buffer to allocate.  The ioctl() 
	 * will not return an error if the buffer is too small, it will
	 * just give us everything that will fit in the buffer.  So, we
	 * have to get the interface information at least twice. 
	 * 
	 * The first time we get the interface info with a buffer of the
	 * legnth of one hundred ifreq structures. A guess.  Each time
	 * after, we increase the buffer size to fit ten more ifreq
	 * structures.  If we get a larger length buffer the last buffer
	 * wasn't large enough and we go through another iteration,
	 * otherwise, we were right the last time around and we stop.
	 */
	last_len = 0;
	len = 100 * ifreq_size;
	while(1)  
	{
		if(!(buf = malloc(len)))
		{
			perror("malloc");
			exit(1);
		}
	
		ifc.ifc_buf = buf;
		ifc.ifc_len = len;
		if( ioctl(sockfd, SIOCGIFCONF, &ifc) < 0)
		{
			if(errno == EINVAL || last_len == 0)
			{
				perror("ioctl");
				exit(1);
			}
		}
		else
		{
			if(ifc.ifc_len == last_len)
				break;
			last_len = ifc.ifc_len;
		}
		len += 10 * ifreq_size;
		free(buf);
	}
	
	*sock = sockfd;
	*buf_len = ifc.ifc_len;
	buffer = ifc.ifc_buf;
	return(buffer);
}

int get_interface_info(struct network *n, const char *if_name)
{
	int sockfd, len, found_flag = 0;
	int ifreq_size = sizeof(struct ifreq);
	struct ifreq *ifr;
	void *buf = NULL, *ptr;
	int get_addr(int sockfd, int ioctrl, struct ifreq *ifr);
	
	buf = get_interface_count(buf, &len, &sockfd);

	/*
	 * For each network interface, if the network interface
	 * matches the interface specified in the if_name arg then
	 * we get the interface address, the subnet mask, and we do
	 * some calculations.  We need to get the subnet mask to
	 * find the subnet bits.
	 */
	ptr = buf;
	while(ptr < buf + len)
	{
		ifr = (struct ifreq *) ptr;	
		if( strcmp(ifr->ifr_name, if_name) == 0)
		{	
			/* Host */
			n->host_addr = get_addr(sockfd, SIOCGIFADDR, ifr);

			/* Network */
			n->class = get_class(n->host_addr);
			get_class_info(n, n->class, 0);
			n->addr = n->host_addr & n->mask;

			/* Subnet */
			n->subnet_mask = get_addr(sockfd, SIOCGIFNETMASK, ifr);
			n->subnet_bits = count_on_bits(n->mask^n->subnet_mask);
			
			found_flag = 1;
			break;
		}
		ptr += ifreq_size;	
	}		
	
	free(buf);
	close(sockfd);
	return found_flag ? 0 : -1;
}

int get_addr(int sockfd, int ioctrl, struct ifreq *ifr)
{
	struct ifreq ifrtmp; 
	struct sockaddr_in *sin;

	ifrtmp = *ifr;
	if( ioctl(sockfd, ioctrl, &ifrtmp))
	{
		perror("ioctl");
		exit(1);
	}
	
	sin = (struct sockaddr_in *) &ifrtmp.ifr_addr;
	return denboize(sin);
}

char **get_interface_names(int *if_count)
{
	int sockfd, len, i = 0;
	int ifreq_size = sizeof(struct ifreq);
	struct ifreq *ifr;
	void *buf = NULL, *ptr;
	char **if_list = NULL;

	buf = get_interface_count(buf, &len, &sockfd);

	ptr = buf;
	if_list = (char **) malloc(sizeof(char *) * (len/ifreq_size +1));
	while(ptr < buf + len)
	{
		ifr = (struct ifreq *) ptr;	
		ptr += ifreq_size;	

		if((if_list[i] = (char *) malloc(sizeof(char) 
			* strlen(ifr->ifr_name))) == NULL)
		{
			fprintf(stderr, 
				"get_interface_info(): couldn't malloc\n");
			return(NULL);
		}
		strcpy(if_list[i], ifr->ifr_name);
		i++;
	}
	
	if_list[i] = NULL;
	*if_count = i;
	
	free(buf);
	close(sockfd);
	return(if_list);
}

void free_interface_names(char **if_names, const int if_count)
{
	int i;

	for(i = 0; i < if_count; i++)
		free(if_names[i]);
	free(if_names);
}

/*
 * Set the basic information for the given network class.
 * We also assign a default value for the network address.
 * The default value of the network address is the first
 * network in the given class.
 */
int get_class_info(struct network *n, char class, const int reset_host)
{
	switch(tolower(class))
	{
	case 'a':
      	n->class = 'A';
		n->bits = 8;
		n->addr = 0x01000000;
		n->mask = 0xff000000;
		break;
	case 'b':
		n->class = 'B';
		n->bits = 16;
		n->addr = 0x80000000;
		n->mask = 0xffff0000;
		break;
	case 'c':
		n->class = 'C';
		n->bits = 24;
		n->addr = 0xc0000000;
		n->mask = 0xffffff00;
		break;
	default:
		return -1;
	}

	if(reset_host) 
		n->host_addr = n->addr +1;
	return 0;
}
